/*
 * board/cmx300/fb.c
 *
 * PXA3XX frame buffer configuration and setup for the CompuLab CM-X300 module
 * Options parsing code is taken from Linux at: drivers/video/pxafb.c
 *
 * (C) Copyright 2010
 * Igor Grinberg, CompuLab, grinberg@compulab.co.il
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <lcd.h>
#include "mfp.h"


#define FB_DEBUG	0

#if FB_DEBUG
#define pr_dbg printf
#define spr_dbg sprintf
#else
#define pr_dbg(...)
#define spr_dbg(...)
#endif

#define LCD_BPP	LCD_COLOR8
#define REG_LCCR0_DEFAULT	0x07b008f8
#define REG_LCCR3_DEFAULT	0x0300FF01

/* VGA - LCD outputs connected to a video DAC  */
/*----------------------------------------------------------------------*/
/* you have to set lccr0 and lccr3 (including pcd) */
#define REG_LCCR0_VGA	0x07b008f8
#define REG_LCCR3_VGA	0x0300FF00
#define VGA_PIXEL_CLK	39722		/* in picoseconds */

/* 640x480x16 @ 61 Hz */
vidinfo_t vga_panel_info = {
	.vl_col		= 640,
	.vl_row		= 480,
	.vl_width	= 640,
	.vl_height	= 480,
	.vl_clkp	= CONFIG_SYS_HIGH,
	.vl_oep		= CONFIG_SYS_HIGH,
	.vl_hsp		= CONFIG_SYS_HIGH,
	.vl_vsp		= CONFIG_SYS_HIGH,
	.vl_dp		= CONFIG_SYS_HIGH,
	.vl_bpix	= LCD_BPP,
	.vl_lbw		= 0,
	.vl_splt	= 0,
	.vl_clor	= 0,
	.vl_tft		= 1,
	.vl_hpw		= 40,
	.vl_blw		= 100,
	.vl_elw		= 18,
	.vl_vpw		= 20,
	.vl_bfw		= 15,
	.vl_efw		= 40,
};

/*----------------------------------------------------------------------*/
/* TOPPOLY LCD  */
#ifdef CONFIG_TD035STEE1
/* you have to set lccr0 and lccr3 (including pcd) */
#define REG_LCCR0_TOPPOLY	0x07b008f8
#define REG_LCCR3_TOPPOLY	0x0330FF00
#define TOPPOLY_PIXEL_CLK	40000		/* in picoseconds */

vidinfo_t toppoly_panel_info = {
	.vl_col		= 480,
	.vl_row		= 640,
	.vl_width	= 480,
	.vl_height	= 640,
	.vl_clkp	= CONFIG_SYS_HIGH,
	.vl_oep		= CONFIG_SYS_HIGH,
	.vl_hsp		= CONFIG_SYS_LOW,
	.vl_vsp		= CONFIG_SYS_LOW,
	.vl_dp		= CONFIG_SYS_HIGH,
	.vl_bpix	= LCD_BPP,
	.vl_lbw		= 0,
	.vl_splt	= 0,
	.vl_clor	= 1,
	.vl_tft		= 1,
	.vl_hpw		= 8,
	.vl_blw		= 24,
	.vl_elw		= 8,
	.vl_vpw		= 2,
	.vl_bfw		= 3,
	.vl_efw		= 4,
};
#endif /* CONFIG_TD035STEE1 */
/*----------------------------------------------------------------------*/

#define LCD_TYPE_ENV_VAR_NAME	"displaytype"

enum display_types {
#ifdef CONFIG_TD035STEE1
	TOPPOLY,
#endif
	VGA,
	CUSTOM,
	NONE
};

static int enable_lcd_lines = 0;
static char *lcd_type;
static int pxafb_parse_options(char *options, vidinfo_t *panel_info);

static int custom_panel_prepare(vidinfo_t *panel_info)
{
	char *custom_lcd_str = getenv(lcd_type);
	char str[256];

	if (!custom_lcd_str)
		return 1;

	strncpy(str, custom_lcd_str, 255);
	str[255] = '\0';

	if (pxafb_parse_options(str, panel_info))
		return 1;

	pr_dbg("%s: LCCR0=0x%x, LCCR3=0x%x\n", "custom_panel_prepare",
		panel_info->pxa.reg_lccr0, panel_info->pxa.reg_lccr3);

	return 0;
}

static enum display_types get_lcd_type(void)
{
	lcd_type = getenv(LCD_TYPE_ENV_VAR_NAME);

	if (!lcd_type || strcmp("none", lcd_type) == 0
		      || strcmp("NONE", lcd_type) == 0)
		return NONE;

#ifdef CONFIG_TD035STEE1
	if (strcmp("toppoly", lcd_type) == 0
	 || strcmp("TOPPOLY", lcd_type) == 0)
		return TOPPOLY;
#endif

	if (strcmp("vga", lcd_type) == 0 || strcmp("VGA", lcd_type) == 0)
		return VGA;

	return CUSTOM;
}

static inline uint get_lclk(void)
{
	u8 acsr_hss = (ACSR & ACCR_HSIO_MASK) >> 14;

	return acsr_hss == 0x3 ? 312
	     : acsr_hss == 0x2 ? 208
	     : acsr_hss == 0x1 ? 156 : 104;
}

static inline u8 get_pcd(uint pixclock)
{
	return (get_lclk() * pixclock / (2 * 1000000)) - 1;
}

extern int toppoly_lcd_init(void);

int lcd_board_init(vidinfo_t *panel_info)
{
	vidinfo_t tmp;
	vidinfo_t *ptmp = &tmp;

	switch (get_lcd_type()) {
#ifdef CONFIG_TD035STEE1
	case TOPPOLY:
		toppoly_panel_info.pxa.reg_lccr0 = REG_LCCR0_TOPPOLY;
		toppoly_panel_info.pxa.reg_lccr3 = REG_LCCR3_TOPPOLY;
		toppoly_panel_info.pxa.reg_lccr3 |= get_pcd(TOPPOLY_PIXEL_CLK);
		ptmp = &toppoly_panel_info;
		toppoly_lcd_init();
		enable_lcd_lines = 1;
		break;
#endif
	case CUSTOM:
		tmp.pxa.reg_lccr0 = REG_LCCR0_DEFAULT;
		tmp.pxa.reg_lccr3 = REG_LCCR3_DEFAULT;

		/* prepare the custom panel including PCD */
		if (custom_panel_prepare(&tmp)) {
			ptmp = &vga_panel_info;
			break;
		}
		enable_lcd_lines = 1;
		break;
	case VGA:
		vga_panel_info.pxa.reg_lccr0 = REG_LCCR0_VGA;
		vga_panel_info.pxa.reg_lccr3 = REG_LCCR3_VGA;
		vga_panel_info.pxa.reg_lccr3 |= get_pcd(VGA_PIXEL_CLK);
		enable_lcd_lines = 1;
	default: /* NONE or VGA */
		ptmp = &vga_panel_info;
		break;
	}

	memcpy(panel_info, ptmp, sizeof(struct vidinfo));

	return !enable_lcd_lines;
}

int lcd_mfpr_gpio_setup(int nbits)
{
#ifdef CONFIG_BOARD_LCD_INIT
	if (!enable_lcd_lines)
		return 1;
#endif

	/* bits 72-75 */
	GPDR2 |= (0xf << 8);
	/* clear AF in MFPRs & set to AF1 (F/L/P clks & BIAS) */
	mfpr_write(MFPR_GPIO72, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO73, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO74, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO75, MFPR_DEFAULT | MFP_AF1);

	/* bits 54-57 */
	GPDR1 |= (0xf << 22);
	/* clear AF in MFPRs & set to AF1 (LDD) */
	mfpr_write(MFPR_GPIO54, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO55, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO56, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO57, MFPR_DEFAULT | MFP_AF1);

	if (nbits <= 4)
		return 0;

	/* bits 58-61 */
	GPDR1 |= (0xf << 26);
	/* clear AF in MFPRs & set to AF1 (LDD) */
	mfpr_write(MFPR_GPIO58, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO59, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO60, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO61, MFPR_DEFAULT | MFP_AF1);

	if (nbits <= 8)
		return 0;

	/* bits 62-69 */
	GPDR1 |= (0x3 << 30);
	GPDR2 |= 0x3f;
	/* clear AF in MFPRs & set to AF1 (LDD) */
	mfpr_write(MFPR_GPIO62, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO63, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO64, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO65, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO66, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO67, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO68, MFPR_DEFAULT | MFP_AF1);
	mfpr_write(MFPR_GPIO69, MFPR_DEFAULT | MFP_AF1);

	return 0;
}

static int parse_opt_mode(const char *this_opt, vidinfo_t *panel_info)
{
	const char *name = this_opt+5;
	unsigned int namelen = strlen(name);
	int res_specified = 0, bpp_specified = 0;
	unsigned int xres = 0, yres = 0, bpp = 0;
	int yres_specified = 0;
	int i;
	for (i = namelen-1; i >= 0; i--) {
		switch (name[i]) {
		case '-':
			namelen = i;
			if (!bpp_specified && !yres_specified) {
				bpp = simple_strtoul(&name[i+1], NULL, 0);
				bpp_specified = 1;
			} else
				goto done;
			break;
		case 'x':
			if (!yres_specified) {
				yres = simple_strtoul(&name[i+1], NULL, 0);
				yres_specified = 1;
			} else
				goto done;
			break;
		case '0' ... '9':
			break;
		default:
			goto done;
		}
	}
	if (i < 0 && yres_specified) {
		xres = simple_strtoul(name, NULL, 0);
		res_specified = 1;
	}
done:
	if (res_specified) {
		pr_dbg("%s: overriding resolution: %dx%d\n",
			"parse_opt_mode", xres, yres);
		panel_info->vl_col = panel_info->vl_width = xres;
		panel_info->vl_row = panel_info->vl_height = yres;
	}
	if (bpp_specified)
		switch (bpp) {
		case 1:
			panel_info->vl_bpix = LCD_MONOCHROME;
			break;
		case 2:
			panel_info->vl_bpix = LCD_COLOR2;
			break;
		case 4:
			panel_info->vl_bpix = LCD_COLOR4;
			break;
		case 8:
			panel_info->vl_bpix = LCD_COLOR8;
			break;
		case 16:
			panel_info->vl_bpix = LCD_COLOR16;
			break;
		default:
			printf("Color depth %d is not valid\n", bpp);
			return 1;
		}

	pr_dbg("%s: overriding bit depth: %d\n", "parse_opt_mode", bpp);

	return 0;
}

static int parse_opt(char *this_opt, vidinfo_t *panel_info)
{
	char s[64];
	uint pixclock;

	s[0] = '\0';

	if (!strncmp(this_opt, "mode:", 5)) {
		return parse_opt_mode(this_opt, panel_info);
	} else if (!strncmp(this_opt, "pixclock:", 9)) {
		pixclock = simple_strtoul(this_opt+9, NULL, 0);
		panel_info->pxa.reg_lccr3 &= ~0xff;
		panel_info->pxa.reg_lccr3 |= get_pcd(pixclock);
	} else if (!strncmp(this_opt, "left:", 5)) {
		panel_info->vl_blw = simple_strtoul(this_opt+5, NULL, 0);
		spr_dbg(s, "left: %u\n", panel_info->vl_blw);
	} else if (!strncmp(this_opt, "right:", 6)) {
		panel_info->vl_elw = simple_strtoul(this_opt+6, NULL, 0);
		spr_dbg(s, "right: %u\n", panel_info->vl_elw);
	} else if (!strncmp(this_opt, "upper:", 6)) {
		panel_info->vl_bfw = simple_strtoul(this_opt+6, NULL, 0);
		spr_dbg(s, "upper: %u\n", panel_info->vl_bfw);
	} else if (!strncmp(this_opt, "lower:", 6)) {
		panel_info->vl_efw = simple_strtoul(this_opt+6, NULL, 0);
		spr_dbg(s, "lower: %u\n", panel_info->vl_efw);
	} else if (!strncmp(this_opt, "hsynclen:", 9)) {
		panel_info->vl_hpw = simple_strtoul(this_opt+9, NULL, 0);
		spr_dbg(s, "hsynclen: %u\n", panel_info->vl_hpw);
	} else if (!strncmp(this_opt, "vsynclen:", 9)) {
		panel_info->vl_vpw = simple_strtoul(this_opt+9, NULL, 0);
		spr_dbg(s, "vsynclen: %u\n", panel_info->vl_vpw);
	} else if (!strncmp(this_opt, "hsync:", 6)) {
		if (simple_strtoul(this_opt+6, NULL, 0) == 0) {
			spr_dbg(s, "hsync: Active Low\n");
			panel_info->vl_hsp = CONFIG_SYS_LOW;
		} else {
			spr_dbg(s, "hsync: Active High\n");
			panel_info->vl_hsp = CONFIG_SYS_HIGH;
		}
	} else if (!strncmp(this_opt, "vsync:", 6)) {
		if (simple_strtoul(this_opt+6, NULL, 0) == 0) {
			spr_dbg(s, "vsync: Active Low\n");
			panel_info->vl_vsp = CONFIG_SYS_LOW;
		} else {
			spr_dbg(s, "vsync: Active High\n");
			panel_info->vl_vsp = CONFIG_SYS_HIGH;
		}
	} else if (!strncmp(this_opt, "dpc:", 4)) {
		if (simple_strtoul(this_opt+4, NULL, 0) == 0) {
			spr_dbg(s, "double pixel clock: false\n");
			panel_info->pxa.reg_lccr3 &= ~LCCR3_DPC;
		} else {
			spr_dbg(s, "double pixel clock: true\n");
			panel_info->pxa.reg_lccr3 |= LCCR3_DPC;
		}
	} else if (!strncmp(this_opt, "outputen:", 9)) {
		if (simple_strtoul(this_opt+9, NULL, 0) == 0) {
			spr_dbg(s, "output enable: active low\n");
			panel_info->vl_oep = CONFIG_SYS_LOW;
			panel_info->pxa.reg_lccr3 |= LCCR3_OEP;
		} else {
			spr_dbg(s, "output enable: active high\n");
			panel_info->vl_oep = CONFIG_SYS_HIGH;
			panel_info->pxa.reg_lccr3 &= ~LCCR3_OEP;
		}
	} else if (!strncmp(this_opt, "pixclockpol:", 12)) {
		if (simple_strtoul(this_opt+12, NULL, 0) == 0) {
			spr_dbg(s, "pixel clock polarity: falling edge\n");
			panel_info->pxa.reg_lccr3 |= LCCR3_PCP;
		} else {
			spr_dbg(s, "pixel clock polarity: rising edge\n");
			panel_info->pxa.reg_lccr3 &= ~LCCR3_PCP;
		}
	} else if (!strncmp(this_opt, "color", 5)) {
		panel_info->vl_clor = 1;	/* no such bit on PXA3XX */
	} else if (!strncmp(this_opt, "mono", 4)) {
		panel_info->vl_clor = 0;	/* no such bit on PXA3XX */
	} else if (!strncmp(this_opt, "active", 6)) {
		panel_info->vl_tft = 1;
		panel_info->pxa.reg_lccr0 |= LCCR0_PAS;
	} else if (!strncmp(this_opt, "passive", 7)) {
		panel_info->vl_tft = 0;
		panel_info->pxa.reg_lccr0 &= ~LCCR0_PAS;
	} else if (!strncmp(this_opt, "single", 6)) {
		panel_info->vl_splt = 0;	/* no such bit on PXA3XX */
	} else if (!strncmp(this_opt, "dual", 4)) {
		panel_info->vl_splt = 1;	/* no such bit on PXA3XX */
	} else if (!strncmp(this_opt, "4pix", 4)) {
		panel_info->vl_lbw = 0;	/* no such bit on PXA3XX */
	} else if (!strncmp(this_opt, "8pix", 4)) {
		panel_info->vl_lbw = 1;	/* no such bit on PXA3XX */
	} else {
		printf("%s: unknown option: %s\n", "parse_opt", this_opt);
		return 1;
	}

	if (s[0] != '\0')
		pr_dbg("%s: override %s", "parse_opt", s);

	return 0;
}

static int pxafb_parse_options(char *options, vidinfo_t *panel_info)
{
	char *this_opt;
	int ret;

	if (!options || !*options)
		return 0;

	pr_dbg("%s: options are \"%s\"\n",
		"pxafb_parse_options", options ? options : "null");

	/* could be made table driven or similar?... */
	while ((this_opt = strsep(&options, ",")) != NULL) {
		ret = parse_opt(this_opt, panel_info);
		if (ret)
			return ret;
	}
	return 0;
}

