/*
 * Copyright (C) 2012 CompuLab, Ltd.
 * Authors: Nikita Kiryanov <nikita@compulab.co.il>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <console.h>

#define PROGRESS_CHECKPOINTS	30
#define UNINITIALIZED		-1

static int progress_bar_enabled;
static int checkpoint_counter;
static int task_size;
static int task_done;
static int granularity;

/*
 * progress_bar_switch - Turn the progress bar on or off.
 */
void progress_bar_switch(int on)
{
	progress_bar_enabled = on;
}

/*
 * progress_bar_init - Prints an unfilled progress bar, and sets its internal
 * state for the given task size. Should be invoked before the relevant
 * progress_bar_advance() calls.
 *
 * @tasksize - The amount of "work" to do; its units depend on the task (could
 * be bytes, page numbers, percentages, etc).
 */
void progress_bar_init(unsigned int tasksize)
{
	int i;

	if (!progress_bar_enabled || tasksize == 0)
		return;

	task_size = tasksize;
	task_done = 0;
	checkpoint_counter = 1;
	granularity = (task_size / PROGRESS_CHECKPOINTS) ?: 1;
	printf("\n[=>");
	for (i = 0; i < PROGRESS_CHECKPOINTS; i++)
		printf(" ");

	printf("]\b");
	for (i = 0; i < PROGRESS_CHECKPOINTS; i++)
		printf("\b");
}

/*
 * Turn the progress bar on, init it, and leave it in a state indicated by "on".
 */
void progress_bar_init_switch(unsigned int tasksize, int on)
{
	progress_bar_switch(1);
	progress_bar_init(tasksize);
	progress_bar_switch(on);
}

/*
 * We need to advance if we passed a new checkpoint, or finished the task
 * completely. While it may seem like finishing a task must necessarily mean
 * hitting the last checkpoint this is not always the case (happens when the
 * size of the task is smaller then the number of checkpoints).
 * */
static int need_to_advance(void)
{
	int passed_checkpoint = task_done >= granularity * checkpoint_counter;
	int finished_task = task_done >= task_size;
	int bar_not_filled = checkpoint_counter <= PROGRESS_CHECKPOINTS;

	return (passed_checkpoint || finished_task) && bar_not_filled;

}

/*
 * progress_bar_advance - Advances the progress bar if needed.
 *
 * @delta - The amount of units by which the task was advanced (i.e. if we
 * read x bytes then delta == x). Delta and tasksize in progress_bar_init()
 * should have the same units.
 */
void progress_bar_advance(unsigned int delta)
{
	if (!progress_bar_enabled || checkpoint_counter == UNINITIALIZED)
		return;

	/* Hit new checkpoint(s) */
	task_done += delta;
	while (need_to_advance()) {
		printf("\b=>");
		checkpoint_counter++;
	}

	/* Finish progress bar. Don't do anything until it is restarted. */
	if (task_done >= task_size) {
		checkpoint_counter = UNINITIALIZED;
		printf("\n");
	}
}

/*
 * progress_bar_abort - Do a line break and set the internal state to "do not
 * print anything".
 *
 * This will leave a partially filled progress bar, representing the partially
 * completed task.
 */
void progress_bar_abort(void)
{
	if (!progress_bar_enabled)
		return;

	printf("\n");
	checkpoint_counter = UNINITIALIZED;
}

/*
 * progress_bar_remove - clear the line on which the progress bar is displayed.
 *
 * progress_bar_redraw - Redraw the progress bar in a new line, with the same
 * state as its last update.
 *
 * These functions are meant to be used in cases where a message (such as a
 * warning) should be printed in the middle of the task. You invoke
 * progress_bar_remove() to make room for the message, print it, and then call
 * progress_bar_redraw() to recreate the progress bar with the state it was in
 * before the message was printed. If a simple printf is used to print the
 * message, one can use progress_bar_printf() instead of the above.
 */
void progress_bar_remove(void)
{
	int i;
	/* +4 is for the "[=>" and "]" parts of the progress bar */
	int progress_bar_size = PROGRESS_CHECKPOINTS + 4;

	if (!progress_bar_enabled)
		return;

	for (i = 0; i < progress_bar_size; i++)
		printf("\b");

	for (i = 0; i < progress_bar_size; i++)
		printf(" ");

	for (i = 0; i < progress_bar_size; i++)
		printf("\b");
}

void progress_bar_redraw(void)
{
	int task_done_save = task_done;

	if (!progress_bar_enabled)
		return;

	progress_bar_init(task_size);
	progress_bar_advance(task_done_save);
}

/*
 * printf without messing up the progress bar.
 */
void progress_bar_printf(const char *fmt, ...)
{
	va_list args;

	progress_bar_remove();
	va_start(args, fmt);
	vprintf(fmt, args);
	va_end(args);
	progress_bar_redraw();
}
