/*
  dm9000.c: Version 1.2 12/15/2003

	A Davicom DM9000 ISA NIC fast Ethernet driver for Linux.
	Copyright (C) 1997  Sten Wang

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

  (C)Copyright 1997-1998 DAVICOM Semiconductor,Inc. All Rights Reserved.

V0.11	06/20/2001	REG_0A bit3=1, default enable BP with DA match
	06/22/2001	Support DM9801 progrmming
			E3: R25 = ((R24 + NF) & 0x00ff) | 0xf000
			E4: R25 = ((R24 + NF) & 0x00ff) | 0xc200
		R17 = (R17 & 0xfff0) | NF + 3
			E5: R25 = ((R24 + NF - 3) & 0x00ff) | 0xc200
		R17 = (R17 & 0xfff0) | NF

v1.00			modify by simon 2001.9.5
	                change for kernel 2.4.x

v1.1   11/09/2001	fix force mode bug

v1.2   03/18/2003       Weilun Huang <weilun_huang@davicom.com.tw>:
			Fixed phy reset.
			Added tx/rx 32 bit mode.
			Cleaned up for kernel merge.

--------------------------------------

       12/15/2003       Initial port to u-boot by
       			Sascha Hauer <saschahauer@web.de>

       06/03/2008	Remy Bohmer <linux@bohmer.net>
			- Fixed the driver to work with DM9000A.
			  (check on ISR receive status bit before reading the
			  FIFO as described in DM9000 programming guide and
			  application notes)
			- Added autodetect of databus width.
			- Made debug code compile again.
			- Adapt eth_send such that it matches the DM9000*
			  application notes. Needed to make it work properly
			  for DM9000A.
			- Adapted reset procedure to match DM9000 application
			  notes (i.e. double reset)
			- some minor code cleanups
			These changes are tested with DM9000{A,EP,E} together
			with a 200MHz Atmel AT91SAM92161 core

	23/02/2010	Igor Grinberg <grinberg@compulab.co.il>
			- Adapt the driver to work with multiple DM9000 chips.
			- Use CONFIG_HAS_ETH1 for assigning the second MAC.
			- Introduce DM9000_NO_EEPROM for cards without srom.
			- Add a function for writing the MAC address to DM9000
			  registers.
			Tested with 2 DM9000A on CM-X300 CompuLab modules.

--------------------------------------

TODO: external MII is not functional, only internal at the moment.
*/

#include <common.h>
#include <command.h>
#include <malloc.h>
#include <net.h>
#include <asm/io.h>

#include "dm9000x.h"

/* Board/System/Debug information/definition ---------------- */

/* #define CONFIG_DM9000_DEBUG */

#ifdef CONFIG_DM9000_DEBUG
#define DM9000_DBG(fmt,args...) printf(fmt, ##args)
#define DM9000_DMP_PACKET(func,packet,length)  \
	do { \
		int i; 							\
		printf(func ": length: %d\n", length);			\
		for (i = 0; i < length; i++) {				\
			if (i % 8 == 0)					\
				printf("\n%s: %02x: ", func, i);	\
			printf("%02x ", ((unsigned char *) packet)[i]);	\
		} printf("\n");						\
	} while(0)
#else
#define DM9000_DBG(fmt,args...)
#define DM9000_DMP_PACKET(func,packet,length)
#endif

#define mdelay(n) ({unsigned long msec = (n); while (msec--) udelay(1000); })

struct dm9000_info {
	int	idx;
	uint	io_addr;	/* Register I/O base address */
	uint	io_data;	/* Data I/O address */

	void (*inblk)(uint io_data, void *data_ptr, int count);
	void (*outblk)(uint io_data, volatile void *data_ptr, int count);
	void (*rx_status)(struct dm9000_info *info, u16 *RxStatus, u16 *RxLen);
};

/* function declaration ------------------------------------- */
static int dm9000_eth_init(struct eth_device *eth, bd_t * bd);
static int dm9000_eth_do_init(struct eth_device *eth, bd_t * bd);
static int dm9000_eth_send(struct eth_device *eth, volatile void *, int);
static int dm9000_eth_recv(struct eth_device *eth);
static void dm9000_eth_halt(struct eth_device *eth);
static int dm9000_probe(struct eth_device *eth);
static u16 phy_read(struct eth_device *eth, int);
static void __maybe_unused phy_write(struct eth_device *eth, int, u16);
static void phy_up(struct eth_device *eth);

#if !defined(DM9000_NO_EEPROM) && !defined(CONFIG_AT91SAM9261EK)
static u16 read_srom_word(struct eth_device *eth, int);
#endif

static u8 DM9000_ior(struct eth_device *eth, int);
static void DM9000_iow(struct eth_device *eth, int reg, u8 value);

/* DM9000 network board routine ---------------------------- */

#define DM9000_outb(d,r) ( *(volatile u8 *)r = d )
#define DM9000_outw(d,r) ( *(volatile u16 *)r = d )
#define DM9000_outl(d,r) ( *(volatile u32 *)r = d )
#define DM9000_inb(r) (*(volatile u8 *)r)
#define DM9000_inw(r) (*(volatile u16 *)r)
#define DM9000_inl(r) (*(volatile u32 *)r)

#ifdef CONFIG_DM9000_DEBUG
static void
dump_regs(struct eth_device *eth)
{
	DM9000_DBG("\n");
	DM9000_DBG("NCR   (0x00): %02x\n", DM9000_ior(eth, 0));
	DM9000_DBG("NSR   (0x01): %02x\n", DM9000_ior(eth, 1));
	DM9000_DBG("TCR   (0x02): %02x\n", DM9000_ior(eth, 2));
	DM9000_DBG("TSRI  (0x03): %02x\n", DM9000_ior(eth, 3));
	DM9000_DBG("TSRII (0x04): %02x\n", DM9000_ior(eth, 4));
	DM9000_DBG("RCR   (0x05): %02x\n", DM9000_ior(eth, 5));
	DM9000_DBG("RSR   (0x06): %02x\n", DM9000_ior(eth, 6));
	DM9000_DBG("ISR   (0xFE): %02x\n", DM9000_ior(eth, DM9000_ISR));
	DM9000_DBG("\n");
}
#endif

static void dm9000_outblk_8bit(uint io_data, volatile void *data_ptr, int count)
{
	int i;
	for (i = 0; i < count; i++)
		DM9000_outb((((u8 *) data_ptr)[i] & 0xff), io_data);
}

static void dm9000_outblk_16bit(uint io_data, volatile void *data_ptr, int count)
{
	int i;
	u32 tmplen = (count + 1) / 2;

	for (i = 0; i < tmplen; i++)
		DM9000_outw(((u16 *) data_ptr)[i], io_data);
}
static void dm9000_outblk_32bit(uint io_data, volatile void *data_ptr, int count)
{
	int i;
	u32 tmplen = (count + 3) / 4;

	for (i = 0; i < tmplen; i++)
		DM9000_outl(((u32 *) data_ptr)[i], io_data);
}

static void dm9000_inblk_8bit(uint io_data, void *data_ptr, int count)
{
	int i;
	for (i = 0; i < count; i++)
		((u8 *) data_ptr)[i] = DM9000_inb(io_data);
}

static void dm9000_inblk_16bit(uint io_data, void *data_ptr, int count)
{
	int i;
	u32 tmplen = (count + 1) / 2;

	for (i = 0; i < tmplen; i++)
		((u16 *) data_ptr)[i] = DM9000_inw(io_data);
}

static void dm9000_inblk_32bit(uint io_data, void *data_ptr, int count)
{
	int i;
	u32 tmplen = (count + 3) / 4;

	for (i = 0; i < tmplen; i++)
		((u32 *) data_ptr)[i] = DM9000_inl(io_data);
}

static void dm9000_rx_status_32bit(struct dm9000_info *info, u16 *RxStatus, u16 *RxLen)
{
	u32 tmpdata;

	DM9000_outb(DM9000_MRCMD, info->io_addr);

	tmpdata = DM9000_inl(info->io_data);
	*RxStatus = __le16_to_cpu(tmpdata);
	*RxLen = __le16_to_cpu(tmpdata >> 16);
}

static void dm9000_rx_status_16bit(struct dm9000_info *info, u16 *RxStatus, u16 *RxLen)
{
	DM9000_outb(DM9000_MRCMD, info->io_addr);

	*RxStatus = __le16_to_cpu(DM9000_inw(info->io_data));
	*RxLen = __le16_to_cpu(DM9000_inw(info->io_data));
}

static void dm9000_rx_status_8bit(struct dm9000_info *info, u16 *RxStatus, u16 *RxLen)
{
	DM9000_outb(DM9000_MRCMD, info->io_addr);

	*RxStatus =
	    __le16_to_cpu(DM9000_inb(info->io_data) +
			  (DM9000_inb(info->io_data) << 8));
	*RxLen =
	    __le16_to_cpu(DM9000_inb(info->io_data) +
			  (DM9000_inb(info->io_data) << 8));
}

/*
  Read DM9000 Vendor and Product IDs
*/

static inline u32 read_id(struct eth_device *eth)
{
	u32 id_val;

	id_val = DM9000_ior(eth, DM9000_VIDL);
	id_val |= DM9000_ior(eth, DM9000_VIDH) << 8;
	id_val |= DM9000_ior(eth, DM9000_PIDL) << 16;
	id_val |= DM9000_ior(eth, DM9000_PIDH) << 24;
	return id_val;
}

/*
  Search DM9000 board, allocate space and register it
*/
int
dm9000_probe(struct eth_device *eth)
{
	struct dm9000_info *info = eth->priv;
	u32 id_val = read_id(eth);

	if (id_val == DM9000_ID) {
		printf("dm9000 i/o: 0x%x, id: 0x%x \n", info->io_addr, id_val);
		return 0;
	} else {
		printf("dm9000 not found at 0x%08x id: 0x%08x\n", info->io_addr, id_val);
		return -1;
	}
}

/* General Purpose dm9000 reset routine */
static void
dm9000_reset(struct eth_device *eth)
{
	DM9000_DBG("resetting DM9000\n");

	/* Reset DM9000,
	   see DM9000 Application Notes V1.22 Jun 11, 2004 page 29 */

	/* Step 1: Power internal PHY */
	phy_up(eth);

	/* Step 2: Software reset */
	DM9000_iow(eth, DM9000_NCR, (NCR_LBK_INT_MAC | NCR_RST));

	do {
		DM9000_DBG("resetting the DM9000, 1st reset\n");
		udelay(25); /* Wait at least 20 us */
	} while (DM9000_ior(eth, DM9000_NCR) & 1);

	DM9000_iow(eth, DM9000_NCR, 0);
	DM9000_iow(eth, DM9000_NCR, (NCR_LBK_INT_MAC | NCR_RST)); /* Issue a second reset */

	do {
		DM9000_DBG("resetting the DM9000, 2nd reset\n");
		udelay(25); /* Wait at least 20 us */
	} while (DM9000_ior(eth, DM9000_NCR) & 1);

	/* Check whether the ethernet controller is present */
	if ((DM9000_ior(eth, DM9000_PIDL) != 0x0) ||
	    (DM9000_ior(eth, DM9000_PIDH) != 0x90))
		printf("ERROR: resetting DM9000 -> not responding\n");
}

static void __dm9000_set_mac_addr(struct eth_device *eth, const unsigned char *addr)
{
	int i, oft;

	for (i = 0, oft = 0x10; i < 6; i++, oft++)
		DM9000_iow(eth, oft, addr[i]);
	for (i = 0, oft = 0x16; i < 8; i++, oft++)
		DM9000_iow(eth, oft, 0xff);
}

void dm9000_set_mac_addr(int eth_idx, const unsigned char *addr)
{
	struct eth_device *eth = eth_get_dev_by_index(eth_idx);

	__dm9000_set_mac_addr(eth, addr);
}

/* Initialize dm9000 board
*/

static int
dm9000_eth_init(struct eth_device *eth, bd_t * bd)
{
	int status, i = 0;

	while (i < 3) {
		status = dm9000_eth_do_init(eth, bd);
		if ((status == 0) || (status == -1))
			return status;
		i++;
	}

	printf("could not establish link\n");
	return -1;
}


static int
dm9000_eth_do_init(struct eth_device *eth, bd_t * bd)
{
	int i, oft, lnk;
	u8 io_mode;
	uchar *enetaddr = bd->bi_enetaddr;
	struct dm9000_info *info = eth->priv;

	DM9000_DBG("dm9000_eth_init()\n");

	/* RESET device */
	dm9000_reset(eth);

	if (dm9000_probe(eth) < 0)
		return -1;

	/* Auto-detect 8/16/32 bit mode, ISR Bit 6+7 indicate bus width */
	io_mode = DM9000_ior(eth, DM9000_ISR) >> 6;

	switch (io_mode) {
	case 0x0:  /* 16-bit mode */
		printf("DM9000: running in 16 bit mode\n");
		info->outblk	= dm9000_outblk_16bit;
		info->inblk	= dm9000_inblk_16bit;
		info->rx_status	= dm9000_rx_status_16bit;
		break;
	case 0x01:  /* 32-bit mode */
		printf("DM9000: running in 32 bit mode\n");
		info->outblk	= dm9000_outblk_32bit;
		info->inblk	= dm9000_inblk_32bit;
		info->rx_status	= dm9000_rx_status_32bit;
		break;
	case 0x02: /* 8 bit mode */
		printf("DM9000: running in 8 bit mode\n");
		info->outblk	= dm9000_outblk_8bit;
		info->inblk	= dm9000_inblk_8bit;
		info->rx_status	= dm9000_rx_status_8bit;
		break;
	default:
		/* Assume 8 bit mode, will probably not work anyway */
		printf("DM9000: Undefined IO-mode:0x%x\n", io_mode);
		info->outblk	= dm9000_outblk_8bit;
		info->inblk	= dm9000_inblk_8bit;
		info->rx_status	= dm9000_rx_status_8bit;
		break;
	}

	/* Program operating register, only internal phy supported */
	DM9000_iow(eth, DM9000_NCR, 0x0);
	/* TX Polling clear */
	DM9000_iow(eth, DM9000_TCR, 0);
	/* Less 3Kb, 200us */
	DM9000_iow(eth, DM9000_BPTR, BPTR_BPHW(3) | BPTR_JPT_600US);
	/* Flow Control : High/Low Water */
	DM9000_iow(eth, DM9000_FCTR, FCTR_HWOT(3) | FCTR_LWOT(8));
	/* SH FIXME: This looks strange! Flow Control */
	DM9000_iow(eth, DM9000_FCR, 0x0);
	/* Special Mode */
	DM9000_iow(eth, DM9000_SMCR, 0);
	/* clear TX status */
	DM9000_iow(eth, DM9000_NSR, NSR_WAKEST | NSR_TX2END | NSR_TX1END);
	/* Clear interrupt status */
	DM9000_iow(eth, DM9000_ISR, ISR_ROOS | ISR_ROS | ISR_PTS | ISR_PRS);

#if defined(CONFIG_HAS_ETH1)
	if (info->idx == 1)
		enetaddr = bd->bi_enet1addr;
#endif
	/* Set Node address */
#if !defined(DM9000_NO_EEPROM) && !defined(CONFIG_AT91SAM9261EK)
	for (i = 0; i < 6; i++)
		((u16 *) enetaddr)[i] = read_srom_word(eth, i);
#endif

	if (is_zero_ether_addr(enetaddr) ||
	    is_multicast_ether_addr(enetaddr)) {
		/* try reading from environment */
		char enetvar[32] = "ethaddr";
		u8 i;
		char *s, *e;

#if defined(CONFIG_HAS_ETH1)
		if (info->idx == 1)
			sprintf(enetvar, "eth1addr");
#endif
		s = getenv (enetvar);
		for (i = 0; i < 6; ++i) {
			enetaddr[i] = s ? simple_strtoul (s, &e, 16) : 0;
			if (s)
				s = (*e) ? e + 1 : e;
		}
	}

	printf("MAC: %02x:%02x:%02x:%02x:%02x:%02x\n",
	       enetaddr[0], enetaddr[1], enetaddr[2],
	       enetaddr[3], enetaddr[4], enetaddr[5]);

	/* fill device MAC address registers */
	__dm9000_set_mac_addr(eth, enetaddr);

	/* read back mac, just to be sure */
	for (i = 0, oft = 0x10; i < 6; i++, oft++)
		DM9000_DBG("%02x:", DM9000_ior(eth, oft));
	DM9000_DBG("\n");

	/* Activate DM9000 */
	/* RX enable */
	DM9000_iow(eth, DM9000_RCR, RCR_DIS_LONG | RCR_DIS_CRC | RCR_RXEN);
	/* Enable TX/RX interrupt mask */
	DM9000_iow(eth, DM9000_IMR, IMR_PAR);

	i = 0;
	while (!(phy_read(eth, 1) & 0x20)) {/* auto negotiation complete bit */
		mdelay(10);
		i++;
		if (i == 1000)
			return -2;
	}
	/* see what we've got */
	lnk = phy_read(eth, 17) >> 12;
	printf("operating at ");
	switch (lnk) {
	case 1:
		printf("10M half duplex ");
		break;
	case 2:
		printf("10M full duplex ");
		break;
	case 4:
		printf("100M half duplex ");
		break;
	case 8:
		printf("100M full duplex ");
		break;
	default:
		printf("unknown: %d ", lnk);
		break;
	}
	printf("mode\n");
	return 0;
}

/*
  Hardware start transmission.
  Send a packet to media from the upper layer.
*/
static int
dm9000_eth_send(struct eth_device *eth, volatile void *packet, int length)
{
	int tmo;
	struct dm9000_info *info = eth->priv;

	DM9000_DMP_PACKET("dm9000_eth_send", packet, length);

	DM9000_iow(eth, DM9000_ISR, IMR_PTM); /* Clear Tx bit in ISR */

	/* Move data to DM9000 TX RAM */
	DM9000_outb(DM9000_MWCMD, info->io_addr); /* Prepare for TX-data */

	/* push the data to the TX-fifo */
	(info->outblk)(info->io_data, packet, length);

	/* Set TX length to DM9000 */
	DM9000_iow(eth, DM9000_TXPLL, length & 0xff);
	DM9000_iow(eth, DM9000_TXPLH, (length >> 8) & 0xff);

	/* Issue TX polling command */
	DM9000_iow(eth, DM9000_TCR, TCR_TXREQ);	/* Cleared after TX complete */

	/* wait for end of transmission */
	tmo = get_timer(0) + 5 * CONFIG_SYS_HZ;
	while ( !(DM9000_ior(eth, DM9000_NSR) & (NSR_TX1END | NSR_TX2END)) ||
		!(DM9000_ior(eth, DM9000_ISR) & IMR_PTM) ) {
		if (get_timer(0) >= tmo) {
			printf("transmission timeout\n");
			break;
		}
	}
	DM9000_iow(eth, DM9000_ISR, IMR_PTM); /* Clear Tx bit in ISR */

	DM9000_DBG("transmit done\n\n");
	return 0;
}

/*
  Stop the interface.
  The interface is stopped when it is brought.
*/
static void
dm9000_eth_halt(struct eth_device *eth)
{
	DM9000_DBG("dm9000_eth_halt\n");

	/* RESET device */
#if !defined(DM9000_KEEP_PHY_ON)
	phy_write(eth, 0, 0x8000);		/* PHY RESET */
	DM9000_iow(eth, DM9000_GPR, 0x01);	/* Power-Down PHY */
#endif
	DM9000_iow(eth, DM9000_IMR, 0x80);	/* Disable all interrupt */
	DM9000_iow(eth, DM9000_RCR, 0x00);	/* Disable RX */
}

/*
  Received a packet and pass to upper layer
*/
static int
dm9000_eth_recv(struct eth_device *eth)
{
	u8 rxbyte, *rdptr = (u8 *) NetRxPackets[0];
	u16 RxStatus, RxLen = 0;
	struct dm9000_info *info = eth->priv;

	/* Check packet ready or not, we must check
	   the ISR status first for DM9000A */
	if (!(DM9000_ior(eth, DM9000_ISR) & 0x01)) /* Rx-ISR bit must be set. */
		return 0;

	DM9000_iow(eth, DM9000_ISR, 0x01); /* clear PR status latched in bit 0 */

	/* There is _at least_ 1 package in the fifo, read them all */
	for (;;) {
		DM9000_ior(eth, DM9000_MRCMDX);	/* Dummy read */

		/* Get most updated data,
		   only look at bits 0:1, See application notes DM9000 */
		rxbyte = DM9000_inb(info->io_data) & 0x03;

		/* Status check: this byte must be 0 or 1 */
		if (rxbyte > DM9000_PKT_RDY) {
			DM9000_iow(eth, DM9000_RCR, 0x00);	/* Stop Device */
			DM9000_iow(eth, DM9000_ISR, 0x80);	/* Stop INT request */
			printf("DM9000 error: status check fail: 0x%x\n", rxbyte);
			return 0;
		}

		if (rxbyte != DM9000_PKT_RDY)
			return 0; /* No packet received, ignore */

		DM9000_DBG("receiving packet\n");

		/* A packet ready now  & Get status/length */
		(info->rx_status)(info, &RxStatus, &RxLen);

		DM9000_DBG("rx status: 0x%04x rx len: %d\n", RxStatus, RxLen);

		/* Move data from DM9000 */
		/* Read received packet from RX SRAM */
		(info->inblk)(info->io_data, rdptr, RxLen);

		if ((RxStatus & 0xbf00) || (RxLen < 0x40) || (RxLen > DM9000_PKT_MAX)) {
			if (RxStatus & 0x100) {
				printf("rx fifo error\n");
			}
			if (RxStatus & 0x200) {
				printf("rx crc error\n");
			}
			if (RxStatus & 0x8000) {
				printf("rx length error\n");
			}
			if (RxLen > DM9000_PKT_MAX) {
				printf("rx length too big\n");
				dm9000_reset(eth);
			}
		} else {
			DM9000_DMP_PACKET("dm9000_eth_recv", rdptr, RxLen);

			DM9000_DBG("passing packet to upper layer\n");
			NetReceive(NetRxPackets[0], RxLen);
		}
	}
	return 0;
}

#if !defined(DM9000_NO_EEPROM) && !defined(CONFIG_AT91SAM9261EK)
/*
  Read a word data from SROM
*/
static u16
read_srom_word(struct eth_device *eth, int offset)
{
	DM9000_iow(eth, DM9000_EPAR, offset);
	DM9000_iow(eth, DM9000_EPCR, 0x4);
	mdelay(8);
	DM9000_iow(eth, DM9000_EPCR, 0x0);
	return (DM9000_ior(eth, DM9000_EPDRL) + (DM9000_ior(eth, DM9000_EPDRH) << 8));
}

static void
write_srom_word(struct eth_device *eth, int offset, u16 val)
{
	DM9000_iow(eth, DM9000_EPAR, offset);
	DM9000_iow(eth, DM9000_EPDRH, ((val >> 8) & 0xff));
	DM9000_iow(eth, DM9000_EPDRL, (val & 0xff));
	DM9000_iow(eth, DM9000_EPCR, 0x12);
	mdelay(8);
	DM9000_iow(eth, DM9000_EPCR, 0);
}
#endif

/*
   Read a byte from I/O port
*/
static u8
DM9000_ior(struct eth_device *eth, int reg)
{
	struct dm9000_info *info = eth->priv;

	DM9000_outb(reg, info->io_addr);
	return DM9000_inb(info->io_data);
}

/*
   Write a byte to I/O port
*/
static void
DM9000_iow(struct eth_device *eth, int reg, u8 value)
{
	struct dm9000_info *info = eth->priv;

	DM9000_outb(reg, info->io_addr);
	DM9000_outb(value, info->io_data);
}

/*
   Read a word from phyxcer
*/
static u16
phy_read(struct eth_device *eth, int reg)
{
	u16 val;

	/* Fill the phyxcer register into REG_0C */
	DM9000_iow(eth, DM9000_EPAR, DM9000_PHY | reg);
	DM9000_iow(eth, DM9000_EPCR, 0xc);	/* Issue phyxcer read command */
	udelay(100);				/* Wait read complete */
	DM9000_iow(eth, DM9000_EPCR, 0x0);	/* Clear phyxcer read command */
	val = (DM9000_ior(eth, DM9000_EPDRH) << 8) | DM9000_ior(eth, DM9000_EPDRL);

	/* The read data keeps on REG_0D & REG_0E */
	DM9000_DBG("phy_read(0x%x): 0x%x\n", reg, val);
	return val;
}

/*
   Write a word to phyxcer
*/
static void __maybe_unused
phy_write(struct eth_device *eth, int reg, u16 value)
{
	/* Fill the phyxcer register into REG_0C */
	DM9000_iow(eth, DM9000_EPAR, DM9000_PHY | reg);

	/* Fill the written data into REG_0D & REG_0E */
	DM9000_iow(eth, DM9000_EPDRL, (value & 0xff));
	DM9000_iow(eth, DM9000_EPDRH, ((value >> 8) & 0xff));
	DM9000_iow(eth, DM9000_EPCR, 0xa);	/* Issue phyxcer write command */
	udelay(500);				/* Wait write complete */
	DM9000_iow(eth, DM9000_EPCR, 0x0);	/* Clear phyxcer write command */
	DM9000_DBG("phy_write(reg:0x%x, value:0x%x)\n", reg, value);
}

static void phy_up(struct eth_device *eth)
{
	/* GPIO0 on pre-activate PHY */
	/* REG_1F bit0 activate phyxcer */
	DM9000_iow(eth, DM9000_GPR, 0);
	/* Let GPIO0 output */
	DM9000_iow(eth, DM9000_GPCR, GPCR_GPIO0_OUT);
	/* Enable PHY */
	DM9000_iow(eth, DM9000_GPR, 0);
}

static uint dev_count = 0;

int dm9000_eth_initialize(uint io_addr, uint io_data)
{
	struct eth_device* dev;
	struct dm9000_info *info;
	u32 id_val;

	dev = (struct eth_device*) malloc(sizeof *dev);
	if (!dev)
		return -1;

	memset(dev, 0, sizeof *dev);

	info = (struct dm9000_info*) malloc(sizeof *info);
	if (!info) {
		free(dev);
		return -1;
	}
	memset(info, 0, sizeof *info);

	info->idx = dev_count;
	info->io_addr = io_addr;
	info->io_data = io_data;

	sprintf(dev->name, "dm9000.%d", dev_count);
	dev->iobase = io_addr;
	dev->init = dm9000_eth_init;
	dev->send = dm9000_eth_send;
	dev->recv = dm9000_eth_recv;
	dev->halt = dm9000_eth_halt;
	dev->priv = info;

	/* Check the device is really present by trying to read the device ID */
	id_val = read_id(dev);
	if (id_val != DM9000_ID) {
		free(info);
		free(dev);

		return -1;
	}

	eth_register(dev);
	dev_count++;

#ifdef DM9000_KEEP_PHY_ON
	phy_up(dev);
#endif

	return 0;
}

